@@ -44,7 +44,7 @@ int getMaxStringLen(const SEXP *col, const int64_t n) {
 int getMaxCategLen(SEXP col) {
   col = getAttrib(col, R_LevelsSymbol);
   if (!isString(col)) error(_("Internal error: col passed to getMaxCategLen is missing levels"));
-  return getMaxStringLen( STRING_PTR(col), LENGTH(col) );
+  return getMaxStringLen( STRING_PTR_RO(col), LENGTH(col) );
 }
 
 const char *getCategString(SEXP col, int64_t row) {
@@ -199,9 +199,8 @@ SEXP fwriteR(
       DFcoerced = PROTECT(allocVector(VECSXP, args.ncol));
       protecti++;
       // potentially large if ncol=1e6 as reported in #1903 where using large VLA caused stack overflow
-      SEXP s = PROTECT(allocList(2));
+      SEXP s = PROTECT(LCONS(R_NilValue, allocList(1)));
       // no protecti++ needed here as one-off UNPROTECT(1) a few lines below
-      SET_TYPEOF(s, LANGSXP);
       SETCAR(s, install("format.POSIXct"));
       for (int j=0; j<args.ncol; j++) {
         SEXP column = VECTOR_ELT(DF, j);
@@ -264,7 +263,7 @@ SEXP fwriteR(
       if (xlength(rn)!=2 || INTEGER(rn)[0]==NA_INTEGER) {
         // not R's default rownames c(NA,-nrow)
         if (xlength(rn) != args.nrow)
-           // Use (long long) to cast R_xlen_t to a fixed type to robustly avoid -Wformat compiler warnings, see #5768, PRId64 didnt work on M1
+           // Use (long long) to cast R_xlen_t to a fixed type to robustly avoid -Wformat compiler warnings, see #5768, PRId64 didn't work on M1
           error(_("input has specific integer rownames but their length (%lld) != nrow (%"PRId64")"), (long long)xlength(rn), args.nrow);  // # nocov
         args.rowNames = INTEGER(rn);
         args.rowNameFun = WF_Int32;
